//
// flodo.c: simple network flow sampling tool
//
// i frequently want to know what the top few network flows are... and yet
// there didn't seem to be any tool which was as simple to use as "vmstat"
// or "iostat".  this tool fills that void for me.
//
// with the default options flodo will dump the top 20 flows every 5 seconds.
//
// flodo is quite basic at this point, understanding only ethernet frames,
// ipv4, tcp, and udp.
//
// flodo can't even do reverse DNS and doesn't bother mapping port numbers
// to services or display icmp or ethernet protocol names.
//

// 
// Copyright (c) 2004 Dean Gaudet <dean@arctic.org>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
// 

// $Id: flodo.c,v 1.8 2004/09/07 06:15:23 dean Exp $

// _BSD_SOURCE gets all the nice packet header structures
#define _BSD_SOURCE 1
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <pcap.h>
#if defined(__linux__) || defined(__FreeBSD__)
#include <net/ethernet.h>
#endif
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#if defined(__NetBSD__)
#include <net/if.h>
#include <net/if_ether.h>
#endif

static const char *progname;
static unsigned top_n = 20;
static pcap_t *handle;
static unsigned secs_per_sample = 5;
static unsigned group_layer = 4;
static struct bpf_program filter_program;

static const char *print_ether(const uint8_t *addr)
{
	static char bufs[4][3*6];
	static unsigned cur_buf = 0;
	char *ret;

	ret = bufs[cur_buf];
	++cur_buf;
	if (cur_buf == 4) cur_buf = 0;
	sprintf(ret, "%02x:%02x:%02x:%02x:%02x:%02x",
		addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
	return ret;
}


static const char *print_ipaddr(uint32_t addr)
{
	static char bufs[4][16];
	static unsigned cur_buf = 0;
	char *ret;

	ret = bufs[cur_buf];
	++cur_buf;
	if (cur_buf == 4) cur_buf = 0;
	addr = ntohl(addr);
	sprintf(ret, "%u.%u.%u.%u",
		(addr >> 24) & 0xff,
		(addr >> 16) & 0xff,
		(addr >> 8) & 0xff,
		addr & 0xff);
	return ret;
}


static const char *print_ipaddr_port(uint32_t addr, uint16_t port)
{
	static char bufs[4][22];
	static unsigned cur_buf = 0;
	char *ret;

	ret = bufs[cur_buf];
	++cur_buf;
	if (cur_buf == 4) cur_buf = 0;
	addr = ntohl(addr);
	sprintf(ret, "%u.%u.%u.%u.%u",
		(addr >> 24) & 0xff,
		(addr >> 16) & 0xff,
		(addr >> 8) & 0xff,
		addr & 0xff,
		ntohs(port));
	return ret;
}


typedef enum {
	ACCUM_UNKNOWN,
	ACCUM_ETHER,
	ACCUM_IP,
	ACCUM_TCP,
	ACCUM_UDP,
} accum_class_t;

typedef struct {
	accum_class_t class;
	union {
		struct ether_header ether;
		struct {
			uint32_t saddr;
			uint32_t daddr;
			uint8_t protocol;
		} ip;
		struct {
			uint32_t saddr;
			uint32_t daddr;
			uint16_t sport;
			uint16_t dport;
		} tcp;
		struct {
			uint32_t saddr;
			uint32_t daddr;
			uint16_t sport;
			uint16_t dport;
		} udp;
	} x;
} accum_t;


static const char *print_accum(const accum_t *a)
{
	static char buf[256];

	switch (a->class) {
	case ACCUM_UNKNOWN:
		snprintf(buf, sizeof(buf), "?????");
		break;

	case ACCUM_ETHER:
		snprintf(buf, sizeof(buf), "ether %s -> %s %04x",
			print_ether(a->x.ether.ether_shost),
			print_ether(a->x.ether.ether_dhost),
			ntohs(a->x.ether.ether_type));
		break;

	case ACCUM_IP:
		snprintf(buf, sizeof(buf), "ip    %-15s -> %-15s %02x",
			print_ipaddr(a->x.ip.saddr),
			print_ipaddr(a->x.ip.daddr),
			a->x.ip.protocol);
		break;

	case ACCUM_TCP:
		snprintf(buf, sizeof(buf), "tcp   %-21s -> %-21s",
			print_ipaddr_port(a->x.tcp.saddr, a->x.tcp.sport),
			print_ipaddr_port(a->x.tcp.daddr, a->x.tcp.dport));
		break;

	case ACCUM_UDP:
		snprintf(buf, sizeof(buf), "udp   %-21s -> %-21s",
			print_ipaddr_port(a->x.udp.saddr, a->x.udp.sport),
			print_ipaddr_port(a->x.udp.daddr, a->x.udp.dport));
		break;

	}
	return buf;
}


static int parse_ip(accum_t *a, const struct ip *ip, size_t len)
{
	struct tcphdr *tcp;
	struct udphdr *udp;

	if (len < sizeof(*ip)) {
		return -1;
	}
	if (group_layer > 3) {
		switch (ip->ip_p) {
		case IPPROTO_TCP:
			if (len >= sizeof(*ip) + sizeof(*tcp)) {
				tcp = (struct tcphdr *)((char *)ip + sizeof(*ip));
				a->class = ACCUM_TCP;
				a->x.tcp.saddr = *(uint32_t *)&ip->ip_src;
				a->x.tcp.daddr = *(uint32_t *)&ip->ip_dst;
				a->x.tcp.sport = tcp->th_sport;
				a->x.tcp.dport = tcp->th_dport;
				return 0;
			}
			break;
		case IPPROTO_UDP:
			if (len >= sizeof(*ip) + sizeof(*udp)) {
				udp = (struct udphdr *)((char *)ip + sizeof(*ip));
				a->class = ACCUM_UDP;
				a->x.udp.saddr = *(uint32_t *)&ip->ip_src;
				a->x.udp.daddr = *(uint32_t *)&ip->ip_dst;
				a->x.udp.sport = udp->uh_sport;
				a->x.udp.dport = udp->uh_dport;
				return 0;
			}
			break;
		}
	}
	a->class = ACCUM_IP;
	a->x.ip.saddr = *(uint32_t *)&ip->ip_src;
	a->x.ip.daddr = *(uint32_t *)&ip->ip_dst;
	a->x.ip.protocol = ip->ip_p;
	return 0;
}


typedef struct accum_hash_str accum_hash_t;
struct accum_hash_str {
	accum_hash_t *next;
	uint64_t total_bytes;
	uint32_t total_packets;
	accum_t a;
};

typedef struct {
	accum_hash_t *head;
	accum_hash_t **tail;
} accum_hashtable_t;

#define HASH_SIZE (2048)
static accum_hash_t *hash[HASH_SIZE];
static accum_hash_t *hash_free_list;
static size_t hash_allocated;
static accum_hash_t **hash_sort_buf;
static size_t hash_sort_buf_allocated;

static void hash_clear(void)
{
	size_t i;
	accum_hash_t **phe;

	// this is really an assumption of hash_key, but i'd
	// rather not test it every hash_key call.
	assert(sizeof(accum_t) % sizeof(uint32_t) == 0);

	for (i = 0; i < HASH_SIZE; ++i) {
		phe = &hash[i];
		if (*phe) {
			while (*phe) {
				phe = &(*phe)->next;
			}
			*phe = hash_free_list;
			hash_free_list = hash[i];
		}
		hash[i] = NULL;
	}
	hash_allocated = 0;
}


static uint32_t hash_key(const accum_t *a)
{
	size_t i;
	uint32_t h = 0;

	for (i = 0; i < sizeof(accum_t)/sizeof(uint32_t); ++i) {
		h = (h << 4) | (h >> 28);
		h ^= *((const uint32_t *)a + i);
	}
	h ^= (h >> 12) ^ (h >> 22);
	return h & (HASH_SIZE - 1);
}


static void hash_accum(const accum_t *a, size_t len)
{
	uint32_t h;
	accum_hash_t **phe;
	accum_hash_t *tmp_he;

	h = hash_key(a);
	phe = &hash[h];
	while (*phe) {
		if (memcmp(&(*phe)->a, a, sizeof(*a)) == 0) {
			// found it...
			tmp_he = *phe;
			tmp_he->total_bytes += len;
			++tmp_he->total_packets;
			// move to front if not already there
			if (tmp_he != hash[h]) {
				*phe = tmp_he->next;
				tmp_he->next = hash[h];
				hash[h] = tmp_he;
			}
			return;
		}
		phe = &(*phe)->next;
	}
	// not found
	if (hash_free_list) {
		tmp_he = hash_free_list;
		hash_free_list = tmp_he->next;
	}
	else {
		tmp_he = malloc(sizeof(*tmp_he));
	}
	tmp_he->total_bytes = len;
	tmp_he->total_packets = 1;
	tmp_he->a = *a;
	tmp_he->next = hash[h];
	hash[h] = tmp_he;
	++hash_allocated;
}


static void pkt_handler(u_char *unused, const struct pcap_pkthdr *header, const u_char *packet)
{
	const struct ether_header *ether;
	accum_t a;

	memset(&a, 0, sizeof(a));
	if (header->len < sizeof(*ether)) {
		a.class = ACCUM_UNKNOWN;
		goto parsed;
	}
	ether = (const struct ether_header *)packet;
	if (group_layer > 2) {
		switch (ntohs(ether->ether_type)) {
		case ETHERTYPE_IP:
			if (parse_ip(&a,
				(const struct ip *)(packet + sizeof(*ether)),
				header->len - sizeof(*ether)) == 0)
				goto parsed;
			break;
		}
	}
	a.class = ACCUM_ETHER;
	a.x.ether = *ether;
parsed:
	hash_accum(&a, header->len);
}


static int hash_compare(const void *va, const void *vb)
{
	const accum_hash_t *a = *(const accum_hash_t **)va;
	const accum_hash_t *b = *(const accum_hash_t **)vb;

	if (a->total_bytes == b->total_bytes) {
		if (a->total_packets == b->total_packets) {
			return memcmp(&a->a, &b->a, sizeof(accum_t));
		}
		else {
			return (int32_t)b->total_packets - (int32_t)a->total_packets;
		}
	}
	return (int64_t)b->total_bytes - (int64_t)a->total_bytes;
}


static void hash_dump(void)
{
	size_t h;
	accum_hash_t *he;
	size_t i;
	uint64_t total_bytes;
	uint32_t total_packets;
	unsigned used_buckets = 0;

	if (hash_allocated > hash_sort_buf_allocated) {
		hash_sort_buf_allocated = 2*hash_allocated;
		hash_sort_buf = realloc(hash_sort_buf, sizeof(hash_sort_buf[0]) * hash_sort_buf_allocated);
	}
	i = 0;
	total_bytes = 0;
	total_packets = 0;
	for (h = 0; h < HASH_SIZE; ++h) {
		he = hash[h];
		if (he) ++used_buckets;
		while (he) {
			total_bytes += he->total_bytes;
			total_packets += he->total_packets;
			hash_sort_buf[i] = he;
			++i;
			he = he->next;
		}
	}
	assert(i == hash_allocated);
	qsort(hash_sort_buf, hash_allocated, sizeof(hash_sort_buf[0]), hash_compare);

	printf ("\n");

	// for debugging the hash function
	if (0 && used_buckets) {
		printf("allocated %zu, used_buckets %u (%.2f%%), avg chain length %.2f\n",
			hash_allocated, used_buckets,
			100.0 * used_buckets / (double)HASH_SIZE,
			hash_allocated / (double)used_buckets);
	}

	printf("%9s %6s %5s %s\n", "bytes/s", "pps", "proto", "details");
	printf("%9.0f %6.0f total\n",
		total_bytes / (double)secs_per_sample,
		total_packets / (double)secs_per_sample);
	for (i = 0; i < hash_allocated && i < top_n; ++i) {
		printf("%9.0f %6.0f %s\n",
			hash_sort_buf[i]->total_bytes / (double)secs_per_sample,
			hash_sort_buf[i]->total_packets / (double)secs_per_sample,
			print_accum(&hash_sort_buf[i]->a));
	}
}


static void alrm_handler(int sig)
{
	pcap_breakloop(handle);
}


static void flodo(void)
{
	struct itimerval itv;
	struct sigaction sa;
	int rc;

	itv.it_interval.tv_usec = 0;
	itv.it_interval.tv_sec = secs_per_sample;
	itv.it_value = itv.it_interval;

        sa.sa_handler = alrm_handler;
        sigemptyset(&sa.sa_mask);
        sigaddset(&sa.sa_mask, SIGALRM);
        sa.sa_flags = 0;
        if (sigaction(SIGALRM, &sa, NULL)) {
                printf("sigaction error: %s\n", strerror(errno));
                exit(1);
        }
        if (setitimer(ITIMER_REAL, &itv, NULL) == -1) {
                printf("setitimer error: %s\n", strerror(errno));
                exit(1);
        }

	setvbuf(stdout, NULL, _IOFBF, BUFSIZ);
	for (;;) {
		hash_clear();
		rc = pcap_loop(handle, -1, pkt_handler, NULL);
		if (rc != -2) break;
		hash_dump();
		fflush(stdout);
		hash_clear();
	}
}


static void usage(void)
{
	printf("usage: %s [options] [libpcap packet filter]\n", progname);
	printf("options:\n");
	printf(" -g group_layer       group the packets at layer 2, 3 or 4\n");
	printf(" -i device            select capture device\n");
	printf(" -p                   enable promiscuous capture (opposite of tcpdump -p)\n");
	printf(" -s N_secs            dump totals every N seconds\n");
	printf(" -t top_N             display only the top N flows\n");
	printf("\nthe packet filtering language is the same as tcpdump(1)\n");
	exit(1);
}


static void pcap_die(const char *errbuf)
{
	fputs(errbuf, stderr);
	fputc('\n', stderr);
	exit(1);
}


static void parse_filter(char **argv, char *dev)
{
	size_t filter_len;
	char *filter_str;
	char *p;
	int i;
	bpf_u_int32 mask;
	bpf_u_int32 net;
	char errbuf[PCAP_ERRBUF_SIZE];

	// need the netmask for dum ipv4 broadcast filtering
	mask = 0;
	if (pcap_lookupnet(dev, &net, &mask, errbuf) == -1) {
		fprintf(stderr, "warning from pcap_lookupnet: %s\n", errbuf);
		// continue anyhow
	}

	// need to paste the arguments together with spaces
	filter_len = 0;
	for (i = optind; argv[i]; ++i) {
		filter_len += strlen(argv[i]) + 1;
	}
	filter_str = malloc(filter_len);
	p = filter_str;
	for (i = optind; argv[i]; ++i) {
		size_t len = strlen(argv[i]);
		memcpy(p, argv[i], len);
		p += len;
		*p++ = ' ';
	}
	p[-1] = '\0';

	if (pcap_compile(handle, &filter_program, filter_str, 1, mask) == -1) {
		fprintf(stderr, "error from pcap_compile: %s\n", pcap_geterr(handle));
		exit(1);
	}
	if (pcap_setfilter(handle, &filter_program) == -1) {
		fprintf(stderr, "error from pcap_setfilter: %s\n", pcap_geterr(handle));
		exit(1);
	}
	free(filter_str);
}



int main(int argc, char **argv)
{
	int c;
	char *dev = NULL;
	char errbuf[PCAP_ERRBUF_SIZE];
	int dl;
	int promiscuous = 0;

	progname = argv[0];

	while ((c = getopt(argc, argv, "g:i:ps:t:")) != -1) {
		switch (c) {
		case 'g':
			group_layer = strtoul(optarg, 0, 0);
			if (group_layer < 2 || group_layer > 4) {
				fprintf(stderr, "group_layer must be 2, 3 or 4\n");
				exit(1);
			}
			break;
		case 'i':
			dev = optarg;
			break;
		case 'p':
			promiscuous = 1;
			break;
		case 's':
			secs_per_sample = strtoul(optarg, 0, 0);
			if (secs_per_sample < 1) {
				fprintf(stderr, "secs_per_sample must be positive\n");
				exit(1);
			}
			break;
		case 't':
			top_n = strtoul(optarg, 0, 0);
			if (top_n < 1) {
				fprintf(stderr, "top_n must be positive\n");
				exit(1);
			}
			break;
		default:
			usage();
		}
	}

	if (dev == NULL) {
		dev = pcap_lookupdev(errbuf);
		if (dev == NULL) pcap_die(errbuf);
	}

	// the to_ms == 50 is necessary for systems which actually do
	// support the timeout ... choosing 50 seems like a good tradeoff
	// on measurement error vs. cpu utilisation.  if you're super
	// anal you could use to_ms == 1 ... but at 50 and default -s 5
	// you'd be looking at .05/5 = 1% error.
	handle = pcap_open_live(dev, 96, promiscuous, 50, errbuf);
	if (handle == NULL) pcap_die(errbuf);

	// verify it's a datalink we can handle...
	switch ((dl = pcap_datalink(handle))) {
	case DLT_EN10MB:
		break;
	default:
		fprintf(stderr, "device %s unknown datalink %s -- %s\n",
			dev, pcap_datalink_val_to_name(dl),
			pcap_datalink_val_to_description(dl));
		exit(1);
	}

	// set up a filter if one is given
	if (argv[optind] != NULL) {
		parse_filter(argv, dev);
	}

	flodo();

	return 0;
}
